;; Generated package description from geiser-racket.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "geiser-racket" "0.16.0.20260727.3" "Support for Racket in Geiser" '((emacs "26.1") (geiser "0.16")) :commit "1d52c67a1bd70b879e62e7320ccf164ff6365ce0" :authors '(("Jose Antonio Ortega Ruiz" . "(jao@gnu.org)")) :maintainer '("Jose Antonio Ortega Ruiz" . "(jao@gnu.org)") :keywords '("languages" "racket" "scheme" "geiser") :url "https://gitlab.com/emacs-geiser/racket")
